# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Plugin module that is used to generate interval information from each tag/datapoint """

import os
import copy
import logging
import datetime
import re

from foglamp.common import logger
import filter_ingest

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2023 Dianomic Systems Inc."
__license__ = "Apache 2.0"

_LOGGER = logger.setup(__name__, level=logging.WARN)

PLUGIN_NAME = "changedetection"

FOGLAMP_DATA_DIR = os.path.join(os.getenv("FOGLAMP_ROOT"), "data")

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Filter that creates interval information for each tag/datapoint',
        'type': 'string',
        'default': PLUGIN_NAME,
        'readonly': 'true'
    },
    'tagNamesRegex': {
        'description': 'Python regular expression to match tag/datapoint names to be processed (special characters like the dot must be escaped)',
        'type': 'string',
        'default': 'sinusoid|random',
        'displayName': 'Regular expression to match datapoint names',
        'order': '2'
    },
    'outputAssetName': {
        'description': 'Output asset name',
        'type': 'string',
        'default': 'DatapointValueIntervals',
        'displayName': 'Output asset name',
        'order': '3'
    },
    'maxReportingInterval': {
        'description': 'Maximum time interval (in secs) after which the summarized information must be forwarded even if no change',
        'type': 'integer',
        'default': '3600',
        'displayName': "Maximum reporting interval (in secs)",
        'minimum': '1',
        'order': "4"
    },
    'forwardData': {
        'description': 'Whether to forward original readings',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Forward original readings',
        'order': '5'
    },
    'enable': {
        'description': 'Enable/Disable filter plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "6"
    }
}


def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        'name': PLUGIN_NAME,
        'version': '3.0.0-1-g409d3e8',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


# Python 3.6 and 3.7 differ in the format of tz offset in datetime strings, handle the same
old_strptime_behavior = True
s = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f%z')
if s[-3] == ':':
    old_strptime_behavior = False


def str_to_datetime(time_str, format):
    """ Convert string representation of datetime string to datetime object

    Args:
        time_str:   datetime string to be converted
        format:     strptime format string
    Returns:
        d:      converted datetime object
    """
    global old_strptime_behavior

    if old_strptime_behavior:
        time_str = time_str[0:-3]+time_str[-2:]

    d = datetime.datetime.strptime(time_str, format)
    return d


def check_regex(s):
    """ Check whether input string is a valid regular expression
    Args:
        s: Input string to check 
    """

    try:
        re.compile(s)
    except Exception: 
        return False
    
    return True


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config:     JSON configuration document for the Filter plugin configuration category
        ingest_ref: filter ingest reference
        callback:   filter callback
    Returns:
        data:       JSON object to be used in future calls to the plugin
    Raises:
    """

    _LOGGER.info("changedetection filter: plugin_init called")

    handle = copy.deepcopy(config)
    handle['callback'] = callback
    handle['ingestRef'] = ingest_ref
    handle['forwardData'] = (config['forwardData']['value'] == 'true')
    handle['maxReportingInterval'] = int(config['maxReportingInterval']['value'])
    handle['state_info'] = {}

    if not check_regex(handle['tagNamesRegex']['value']):
        _LOGGER.fatal("Invalid regular expression given for matching tag names: {}".format(handle['tagNamesRegex']['value']))
        _LOGGER.fatal("Disabling filter {}".format(PLUGIN_NAME))
        handle['enable']['value'] == 'false'

    return handle


def generate_reading(tag, interval_data, timestamp, outputAssetName):
    """ Generate reading for a closed interval

    Args:
        tag:            Tag/Datapoint name
        interval_data:  Time interval information where a given tag continued to have a certain value
        timestamp:      Reading timestamp
        outputAssetName: Asset name to put into the reading
    Returns:
        rdng:           The reading generated with the input data
    """
    rdngData = {'tag': tag,
                'value': interval_data['value'], 
                'start_time': interval_data['start_time'],
                'end_time': interval_data['end_time'] }
    
    rdng = {'asset': outputAssetName,
            'readings': rdngData,
            'timestamp': timestamp
            }
    
    return rdng


def generate_readings_for_all_intervals(handle):
    """ Generate readings for all known closed intervals

    Args:
        handle:     handle returned by the plugin initialisation call
    Returns:
        closed_interval_readings: An array of readings corresponding to all known closed intervals
    """
    closed_interval_readings = []
    # Process all known intervals and generate readings and ingest them
    for tag in handle['state_info'].keys():
        # for open intervals, copy 'start_time' into 'end_time'
        if 'end_time' not in handle['state_info'][tag]:
            handle['state_info'][tag]['end_time'] = handle['state_info'][tag]['start_time']
            _LOGGER.debug("generate_readings_for_all_intervals: interval has not been closed, set end_time=start_time: {}".format(handle['state_info'][tag]))

        rdng = generate_reading(tag, handle['state_info'][tag],
                                handle['state_info'][tag]['end_time'], handle['outputAssetName']['value'])
        _LOGGER.info("{}: generate_readings_for_all_intervals(): {}".format(PLUGIN_NAME, rdng))
        closed_interval_readings.append(rdng)

    return closed_interval_readings


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    Args:
        handle:     handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.debug("changedetection filter: plugin_reconfigure called")

    _LOGGER.info("Old config {} \n new config {} for {} plugin ".format(handle, new_config, PLUGIN_NAME))
    new_handle = plugin_init(new_config, handle['ingestRef'], handle['callback'])

    # if filter has been disabled via reconf, send readings for all known closed intervals
    if handle['enable']['value'] == 'true' and new_handle['enable']['value'] == 'false':
        closed_interval_readings = generate_readings_for_all_intervals(handle)
        if len(closed_interval_readings):
            filter_ingest.filter_ingest_callback(new_handle['callback'], new_handle['ingestRef'], closed_interval_readings)
    
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
    """
    closed_interval_readings = generate_readings_for_all_intervals(handle)
    if len(closed_interval_readings):
        filter_ingest.filter_ingest_callback(handle['callback'], handle['ingestRef'], closed_interval_readings)

    handle['callback'] = None
    handle['ingestRef'] = None

    _LOGGER.info('{} filter plugin shutdown.'.format(PLUGIN_NAME))


def plugin_ingest(handle, data):
    """ Modify readings data and pass it onward

    Args:
        handle: handle returned by the plugin initialisation call
        data:   readings data
    """

    _LOGGER.debug("changedetection filter: plugin_ingest called")

    if handle['enable']['value'] == 'false':
        # Filter not enabled, nothing to do here, just forward data down the filter pipeline
        filter_ingest.filter_ingest_callback(handle['callback'],  handle['ingestRef'], data)
        return

    # Filter is enabled: compute for each reading
    closed_interval_readings = []
    for element in data: # process each reading
        _LOGGER.debug("1. changedetection filter: plugin_ingest: Handling reading: {}".format(element))
        # process each DP within current reading
        for tag,val in element['readings'].items():
            if not re.fullmatch(handle['tagNamesRegex']['value'], tag):
                _LOGGER.debug("2. Not processing tag {} as it doesn't match regex for tag names: {}".format(tag, handle['tagNamesRegex']['value']))
                continue

            timestamp = element['timestamp'] if 'timestamp' in element else element['ts']
            if tag not in handle['state_info']:
                handle['state_info'][tag] = { 'value': val, 'start_time': timestamp }
                _LOGGER.debug("3. changedetection filter: plugin_ingest: No open interval for tag {}, created new interval: {}"
                                .format(tag, handle['state_info'][tag]))

            else: # if an interval already exists for this tag, update end_time & check whether the value has changed
                handle['state_info'][tag]['end_time'] = timestamp
                _LOGGER.debug("4. changedetection filter: plugin_ingest: Updated end_time for tag {} interval data: {}"
                                .format(tag, handle['state_info'][tag]))

                start_time = str_to_datetime(handle['state_info'][tag]['start_time'], '%Y-%m-%d %H:%M:%S.%f%z')
                end_time = str_to_datetime(handle['state_info'][tag]['end_time'], '%Y-%m-%d %H:%M:%S.%f%z')
                time_diff = end_time - start_time

                # if tag value has changed or reporting interval has elapsed, close current interval and start a new one
                if (handle['state_info'][tag]['value'] != val) or \
                   (time_diff >= datetime.timedelta(seconds=handle['maxReportingInterval'])):
                    if handle['state_info'][tag]['value'] != val:
                        _LOGGER.debug("5. changedetection filter: plugin_ingest: tag {} value has changed from {} to {}"
                                        .format(tag, handle['state_info'][tag]['value'], val))
                    else:
                        _LOGGER.debug("6. changedetection filter: plugin_ingest: Generating reading since the interval has been open for {} secs, i.e. more than reporting interval {} secs"
                                        .format(time_diff.total_seconds(), handle['maxReportingInterval']))
                        
                    rdng = generate_reading(tag, handle['state_info'][tag], timestamp, handle['outputAssetName']['value'])
                    _LOGGER.info("7. changedetection filter: plugin_ingest: Generated closed interval reading: {}".format(rdng))
                    closed_interval_readings.append(rdng)

                    # start a new interval
                    handle['state_info'][tag] = { 'value': val, 'start_time': timestamp }
                    _LOGGER.debug("8. changedetection filter: plugin_ingest: created new interval for tag {}: {}"
                                .format(tag, handle['state_info'][tag]))

    if handle['forwardData']:
        data.extend(closed_interval_readings)
    else:
        data = closed_interval_readings

    # Pass readings onwards
    if len(data):
        filter_ingest.filter_ingest_callback(handle['callback'],  handle['ingestRef'], data)

    _LOGGER.debug("{} filter ingest done".format(PLUGIN_NAME))
